static char rcsid[] = "@(#)$Id: flowed.c,v 2.4 2013-10-13 07:18:53 hurtta Exp $";

/******************************************************************************
 *  The Elm (ME+) Mail System  -  $Revision: 2.4 $   $State: Exp $
 *
 *  Author: Kari Hurtta <hurtta+elm@siilo.FMI.FI> 
 *      or  Kari Hurtta <elm@elmme-mailer.org>
 *****************************************************************************/


#include "def_melib.h"
#include "s_me.h"

DEBUG_VAR(Debug,__FILE__,"mime");

static char *us2s P_((unsigned char *str));
static char *us2s(str) 
     unsigned char *str;
{
    return (char *)str;
}

/* RFC 2646:   The Text/Plain Format and DelSp Parameters  */

struct flowed_line {
    int quote_level;

    enum fl_line_class {
	fl_special,
	fl_fixed,
	fl_flowed,
	fl_too_long
    }             line_class;
    unsigned int   have_nl :1;   /* newline readed (and not stored) */
    unsigned int   eof :1;       /* got EOF (or error)              */
    
    unsigned char     *buffer;
    int buffer_len;
};

#if DEBUG
static const char * line_class_string P_((enum fl_line_class lc));
static const char * line_class_string(lc)
     enum fl_line_class lc;
{
    switch (lc) {
    case fl_special: return "special";
    case fl_fixed:   return "fixed";
    case fl_flowed:  return "flowed";
    case fl_too_long: return "too long";
    }

    return "<unknown>";
}
#endif

static void flowed_line_init P_((struct flowed_line *line));
static void flowed_line_init(line)
     struct flowed_line *line;
{
    line->quote_level = 0;
    line->line_class  = fl_special;
    line->have_nl     = 0;
    line->eof         = 0;

    line->buffer =  NULL;
    line->buffer_len = 0;
}

static void flowed_line_dest P_((struct flowed_line *line));
static void flowed_line_dest(line)
     struct flowed_line *line;
{
    line->quote_level = 0;
    line->line_class  = fl_special;
    line->have_nl     = 0;
    line->eof         = 0;

    if (line->buffer) {
	free(line->buffer);
	line->buffer = NULL;
    }
    line->buffer_len = 0;
}

#define INITIAL_LINE_LEN  80
#define LINE_LEN_STEP    512

/* flowed lines should be under 80 characters */
#define MAX_FLOWED_LINE  1002

/* Returns -1 if EOF (or error seen) before any characters */
int get_new_flowed_line P_((struct flowed_line *line, struct in_state   *state_in,
			    int delsp));
int get_new_flowed_line(line,state_in,delsp)
     struct flowed_line *line; 
     struct in_state   *state_in;
     int delsp;
{
    int alloced = INITIAL_LINE_LEN;
    int ch = EOF;
    int ret = 0;

    line->buffer = safe_realloc(line->buffer,alloced);

    line->quote_level = 0;
    line->line_class  = fl_fixed;
    line->have_nl     = 0;
    line->buffer_len  = 0;
    line->eof         = 0;
    line->buffer[0]   = '\0';

    ch = state_getc(state_in);
    
    if (EOF == ch) {
	ret = -1;
	line->eof         = 1;
	line->line_class  = fl_special;
    
	free(line->buffer);
	line->buffer      = NULL;

	goto out;
    }

    /* Count quote level */
    while ('>' == ch) {
	line->quote_level++;

	ch = state_getc(state_in);
    }

    /* Remove space-stuffing */
    if (' ' == ch)
	ch = state_getc(state_in);	

    while (EOF != ch && line->buffer_len < MAX_FLOWED_LINE) {

	if (line->buffer_len >= alloced) {
	    alloced = ( line->buffer_len / LINE_LEN_STEP + 1) * LINE_LEN_STEP;

	    if (alloced > MAX_FLOWED_LINE+1)
		alloced = MAX_FLOWED_LINE+1;

	    line->buffer = safe_realloc(line->buffer,alloced);
	}

	if (line->buffer_len >= alloced)
	    mime_panic(__FILE__,__LINE__,"get_new_flowed_line",
		       "Bad len");

	if ('\r' == ch) {
	    int peekch = state_getc(state_in);

	    if ('\n' == peekch) {   /* Got newline */
		ch = peekch;

		line->have_nl = 1;
		break;
	    } else if (EOF != peekch)
		state_ungetc(peekch,state_in);

	} else if ('\n' == ch) {

	    line->have_nl = 1;
	    break;
	}

	line->buffer[line->buffer_len++] = ch;
	ch = state_getc(state_in);	
    }

    if (line->buffer_len > 0 && line->buffer_len < MAX_FLOWED_LINE
	&& ' ' == line->buffer[line->buffer_len-1]) {

	/* Usenet Signature Convention */
	if (3 == line->buffer_len && 0 == strncmp("-- ",us2s(line->buffer),3))
	    line->line_class  = fl_special;
	else {
	    line->line_class  = fl_flowed;
	    
	    if (delsp)
		line->buffer_len--;
	}
    } else if (line->buffer_len >= MAX_FLOWED_LINE)
	line->line_class  = fl_too_long;

    ret = line->buffer_len;

    alloced = line->buffer_len+1;
    
    line->buffer = safe_realloc(line->buffer,alloced);
    line->buffer[line->buffer_len] = '\0';

    if (ch == EOF)
	line->eof         = 1;

 out:

    DPRINT(Debug,20,(&Debug, 
		     "get_new_flowed_line=%d, quote level=%d, line class=%s%s",
		     ret,line->quote_level,
		     line_class_string(line->line_class),
		     line->eof     ? ", got EOF" : ""));

    if (line->buffer) {
	DPRINT(Debug,25,(&Debug,", buffer="));
	DEBUG_PRINT_BUFFER(Debug,25,line->buffer_len,line->buffer);
    }

    DPRINT(Debug,20,(&Debug, "\n"));
		     		   
    return ret;
}


static int collect_special P_((mime_t *body, struct in_state   *state_in, struct out_state  *state_out,
			       const struct decode_opts *decode_opt,struct flowed_line *current_line,
			       int delsp));
static int collect_special(body,state_in,state_out,decode_opt,current_line,delsp)
     mime_t *body; 
     struct in_state   *state_in; 
     struct out_state  *state_out;
     const struct decode_opts *decode_opt;
     struct flowed_line *current_line;
     int delsp;
{
    int r = 0;
    int last_level = current_line->quote_level;
    struct decode_opts  DECODE_OPT = *decode_opt;

    struct pager_range * decode_data = NULL;


    if (last_level > 0) {
	decode_data =  state_add_simple_pager_range(state_out,decode_opt->range,0,last_level);

	DPRINT(Debug,15,(&Debug,
			 "collect_special: decode_data quote_level=%d range_flags=%d\n",
			 get_pager_range_quote_level(decode_data),
			 get_pager_range_flags(decode_data)));

	DECODE_OPT.range  =  decode_data;
    }

    while (fl_special == current_line->line_class &&
	   last_level == current_line->quote_level &&
	   r >= 0) {
	int i;

	state_add_prefix(state_out,&DECODE_OPT);

	for (i = 0; i < current_line->buffer_len; i++)
	    state_putc(current_line->buffer[i],state_out);

	if (current_line->have_nl)
	    state_putc('\n',state_out);

	r = get_new_flowed_line(current_line,state_in,delsp);
    }

    if (decode_data)
	free_pager_range(&decode_data);

    return r;
}

static int collect_fixed P_((mime_t *body, struct in_state   *state_in, struct out_state  *state_out,
			     const struct decode_opts *decode_opt,struct flowed_line *current_line,
			     int delsp));
static int collect_fixed(body,state_in,state_out,decode_opt,current_line,delsp)
     mime_t *body; 
     struct in_state   *state_in; 
     struct out_state  *state_out;
     const struct decode_opts *decode_opt;
     struct flowed_line *current_line;
     int delsp;
{
    /* Almost as collect_special */

    int r = 0;
    int last_level = current_line->quote_level;
    struct decode_opts  DECODE_OPT = *decode_opt;

    struct pager_range * decode_data =  
	state_add_simple_pager_range(state_out,decode_opt->range,PR_PREFORMAT,
				     last_level);

    DPRINT(Debug,15,(&Debug,
		     "collect_fixed: decode_data quote_level=%d range_flags=%d\n",
		     get_pager_range_quote_level(decode_data),
		     get_pager_range_flags(decode_data)));
    
    DECODE_OPT.range  =  decode_data;

    while (fl_fixed == current_line->line_class &&
	   last_level == current_line->quote_level &&
	   r >= 0) {
	int i;

	state_add_prefix(state_out,&DECODE_OPT);

	for (i = 0; i < current_line->buffer_len; i++)
	    state_putc(current_line->buffer[i],state_out);

	if (current_line->have_nl)
	    state_putc('\n',state_out);

	r = get_new_flowed_line(current_line,state_in,delsp);
    }

    free_pager_range(&decode_data);

    return r;

}

static int collect_flowed P_((mime_t *body, struct in_state   *state_in, struct out_state  *state_out,
			      const struct decode_opts *decode_opt,struct flowed_line *current_line,
			      int delsp));
static int collect_flowed(body,state_in,state_out,decode_opt,current_line,delsp)
     mime_t *body; 
     struct in_state   *state_in; 
     struct out_state  *state_out;
     const struct decode_opts *decode_opt;
     struct flowed_line *current_line;
     int delsp;
{
    /* flowed paragraph   consists
       current_line->line_class  == fl_flowed lines followed by one
       current_line->line_class  == fl_fixed line */


    int r = 0;
    int last_level = current_line->quote_level;
    struct decode_opts  DECODE_OPT = *decode_opt;

    struct pager_range * decode_data =  
	state_add_simple_pager_range(state_out,decode_opt->range,
				     PR_WORD_WRAP | PR_JOIN_LINES, last_level);

    DPRINT(Debug,15,(&Debug,
		     "collect_flowed: decode_data quote_level=%d range_flags=%d\n",
		     get_pager_range_quote_level(decode_data),
		     get_pager_range_flags(decode_data)));
    
    DECODE_OPT.range  =  decode_data;

    /* 1) Flowed lines */

    while (fl_flowed == current_line->line_class &&
	   last_level == current_line->quote_level &&
	   r >= 0) {
	int i;

	state_add_prefix(state_out,&DECODE_OPT);

	for (i = 0; i < current_line->buffer_len; i++)
	    state_putc(current_line->buffer[i],state_out);

	if (current_line->have_nl)
	    state_putc('\n',state_out);

	r = get_new_flowed_line(current_line,state_in,delsp);
    }

    /* 2) and fixed line */

    if (fl_fixed == current_line->line_class &&
	last_level == current_line->quote_level &&
	r >= 0) {
	int i;
	
	state_add_prefix(state_out,&DECODE_OPT);
	
	for (i = 0; i < current_line->buffer_len; i++)
	    state_putc(current_line->buffer[i],state_out);
	
	if (current_line->have_nl)
	    state_putc('\n',state_out);
	
	r = get_new_flowed_line(current_line,state_in,delsp);
    }
    

    free_pager_range(&decode_data);

    return r;
}

static int collect_overflow P_((mime_t *body, struct in_state   *state_in, struct out_state  *state_out,
				const struct decode_opts *decode_opt,struct flowed_line *current_line,
				int delsp));
static int collect_overflow(body,state_in,state_out,decode_opt,current_line,delsp)
     mime_t *body; 
     struct in_state   *state_in; 
     struct out_state  *state_out;
     const struct decode_opts *decode_opt;
     struct flowed_line *current_line;
     int delsp;
{
    /* Almost as collect_special */

    int r = 0;
    int last_level = current_line->quote_level;
    struct decode_opts  DECODE_OPT = *decode_opt;

    struct pager_range * decode_data = NULL;


    if (last_level > 0) {
	decode_data =  state_add_simple_pager_range(state_out,decode_opt->range,0,last_level);

	DPRINT(Debug,15,(&Debug,
			 "collect_overflow: decode_data quote_level=%d range_flags=%d\n",
			 get_pager_range_quote_level(decode_data),
			 get_pager_range_flags(decode_data)));

	DECODE_OPT.range  =  decode_data;
    }

    while (fl_too_long == current_line->line_class &&
	   last_level == current_line->quote_level &&
	   r >= 0) {
	int i;

	state_add_prefix(state_out,&DECODE_OPT);

	for (i = 0; i < current_line->buffer_len; i++)
	    state_putc(current_line->buffer[i],state_out);

	if (current_line->have_nl)   /* Should not happen */
	    state_putc('\n',state_out);
	else {
	    int ch = EOF;

	    /* Copy rest of line, which was not readed because of overflow 
	     * This does not check is line flowed or fixed
	     */

	    do {

		ch = state_getc(state_in);
		
		if (EOF == ch) 
		    break;
	    
		state_putc(ch,state_out);
	    } while ('\n' != ch);
	}

	r = get_new_flowed_line(current_line,state_in,delsp);
    }

    if (decode_data)
	free_pager_range(&decode_data);

    return r;
}


void flowed_text_decoder(body,state_in,state_out,decode_opt)
     mime_t *body;
     struct in_state   *state_in;
     struct out_state  *state_out;
     const struct decode_opts *decode_opt;
{
    const char *pv = get_mime_param_compat(body->TYPE_opts,"delsp");
    int delsp = 0;
    int r;
    int errors2 UNUSED_VAROK;
    int overflow = 0;

    struct flowed_line current_line;

    DPRINT(Debug,15,(&Debug,"flowed_text_decoder: "));
    if (pv) {
	DPRINT(Debug,15,(&Debug,"delsp=%s",pv));

	delsp = 0 == istrcmp(pv,"yes");
    }
    DPRINT(Debug,15,(&Debug,"\n"));

    flowed_line_init(&current_line);

    r = get_new_flowed_line(&current_line,state_in,delsp);

    while (r >= 0) {
	int errors;
	enum fl_line_class prev UNUSED_VAROK = 
	    current_line.line_class;

	switch (current_line.line_class) {
	    
	case fl_special:
	    
	    r = collect_special(body,state_in,state_out,decode_opt,&current_line,delsp);
	    
	    break;
	    
	case fl_fixed:
	    
	    r = collect_fixed(body,state_in,state_out,decode_opt,&current_line,delsp);
	    
	    break;
	    
	case fl_flowed:
	    
	    r = collect_flowed(body,state_in,state_out,decode_opt,&current_line,delsp);
	    
	    break;
	    
	case fl_too_long:
	    
	    if (!overflow) {
		struct pager_range *title_range = 
		    state_add_simple_pager_range(state_out,NULL,PR_MAX_WIDTH,0);

		
		/* \n resets this */
		set_out_state_line_mode(state_out,pg_BOLD,title_range); 
		state_printf(state_out,
			     CATGETS(elm_msg_cat, MeSet, 
				     MeDecodeTooLongLines,
				     "[ text/plain format with too long lines, treating like fixed format. ]\n"));

		free_pager_range(&title_range);

		overflow = 1;
	    }

	    r = collect_overflow(body,state_in,state_out,decode_opt,&current_line,delsp);
	    
	    break;
	    
	default:
	    mime_panic(__FILE__,__LINE__,"flowed_text_decoder",
		       "Bad current_line.line_class");
	    
	}

	errors = print_in_errors(state_in,state_out,decode_opt);

	DPRINT(Debug,10,(&Debug, 
			 "flowed_text_decoder: %d errors when reading %s, block return %d; next block %s\n",
			 errors,line_class_string(prev),r, line_class_string(current_line.line_class)));

	/* Retry after errors */
	if (errors > 0 && current_line.buffer_len < 1 && r < 0)
	    r = get_new_flowed_line(&current_line,state_in,delsp);

    }

    flowed_line_dest(&current_line);

    errors2 = print_in_errors(state_in,state_out,decode_opt);
    
    DPRINT(Debug,10,(&Debug, 
		     "flowed_text_decoder: %d errors on end\n",errors2));
}

/*
 * Local Variables:
 *  mode:c
 *  c-basic-offset:4
 *  buffer-file-coding-system: iso-8859-1
 * End:
 */
